
import jax 
import jax.numpy as jnp

from functools import partial


@jax.jit
def find_ancestors_jax(mask, node):
    """Find ancestors of a node in a graph.

    Args:
        mask (Array): Adjacency matrix of a directed graph.
        node (int): Node of interest.

    Returns:
        _type_: _description_
    """
    num_nodes = mask.shape[0]
    is_ancestor = jnp.zeros(num_nodes, dtype=jnp.bool_)
    stack = jnp.empty(num_nodes, dtype=jnp.int32)
    stack = stack.at[0].set(node)
    
    def body_fn(carry, i):
        is_ancestor, stack = carry
        current_node = stack[i]
        current_parents = mask[current_node, :]
        
        def inner_body_fn(carry, j):
            is_ancestor, stack = carry
            value = current_parents[j]
            cond = value & (j != current_node) & (~is_ancestor[j])
            
            def true_fn(is_ancestor, stack):
                is_ancestor = is_ancestor.at[j].set(True)
                stack = stack.at[i+1].set(j)
                return is_ancestor, stack
            def false_fn(is_ancestor, stack):
                return is_ancestor, stack
            
            is_ancestor, stack = jax.lax.cond(cond, true_fn, false_fn, is_ancestor, stack)
            return (is_ancestor, stack), None
        
        (is_ancestor, stack), _ = jax.lax.scan(inner_body_fn, (is_ancestor, stack), jnp.arange(num_nodes))
        return (is_ancestor, stack), None
    
    (is_ancestor, stack), _ = jax.lax.scan(body_fn, (is_ancestor, stack), jnp.arange(num_nodes))
    

    return is_ancestor



@partial(jax.jit, static_argnums=(2,))
def faithfull_mask(base_mask, condition_mask, conditioned_nodes="unchanged"):
    """ Faithfull mask update for conditioning"""
    
    graph = base_mask.astype(jnp.bool_).copy()
    base_mask = base_mask.astype(jnp.bool_) # Rows are paraents, columns are children
    condition_mask = condition_mask.astype(jnp.bool_)
    num_nodes = base_mask.shape[0]
    
    def body_fn(carry, i):
        base_mask, condition_mask = carry
        
        def condition_case(base_mask, condition_mask):
            # We need to update all ancestors of i
            is_ancestor = find_ancestors_jax(graph, i)
            is_ancestor = is_ancestor & (~condition_mask)
            all_ancestors = jnp.nonzero(is_ancestor, size=num_nodes, fill_value=i)[0]
            # They will now depend on i
            base_mask = base_mask.at[all_ancestors,i].set(True)
            # They will now depend on each other!
            base_mask = base_mask | (is_ancestor[:,None] & is_ancestor[None,:])
            # The parents of all children of i will now depend on each other
            children_of_i = base_mask[:,i]
            parents_of_children_of_i = base_mask & (children_of_i[:,None] & ~children_of_i[None,:])
            parents_of_children_of_i = jnp.any(parents_of_children_of_i, axis=0)

            return base_mask, condition_mask
        
        def uncondition_case(base_mask, condition_mask):
            return base_mask, condition_mask
        
        base_mask, condition_mask = jax.lax.cond(condition_mask[i], condition_case, uncondition_case, base_mask, condition_mask)
        

        return (base_mask, condition_mask), None

    (base_mask, condition_mask), _ = jax.lax.scan(body_fn, (base_mask, condition_mask), jnp.arange(num_nodes))
        
    
    return base_mask



@partial(jax.jit, static_argnums=(2,3))
def min_faithfull_mask(mask, condition_mask, top_mode=0, conditioned_nodes="unchanged"):
    """ Minimally faithfull mask update for conditioning"""
    num_nodes = mask.shape[0]
    I = moralize(mask)
    H = jnp.zeros_like(mask, dtype=jnp.bool_)
    # 0 is child, 1 is parent
    UPSTREAM = top_mode
    DOWNSTREAM = 1 - top_mode
    num_parents_or_childs = jnp.sum(mask & (~condition_mask[None, :] & ~condition_mask[:, None]), axis=UPSTREAM)
    #print(num_parents_or_childs)
    S = (num_parents_or_childs == 1) & (~condition_mask) # Frontier set
    M = jnp.zeros((num_nodes), dtype=jnp.bool_) # Marked nodes

    def cond_fn(val):
        S, _, _, _ = val
        return jnp.any(S) 
    
    def body_fn(val):
        S, M, I, H = val
        #print(S)
        # Find the node with the fewest edges added
        v = min_fill_heuristic(mask, I, S,M, top_mode)
        # print("Frontal set: ",S)
        # print("Marked: ", M)
        # print("Selected: ",v)
        # Add edge in I between unmarked neighbours in I 
        neighbours_v = I[v,:] & (~M)
        I = I | (neighbours_v[:,None] & neighbours_v[None,:])
        # Make unmarked neighbours of v, the parents of v in H
        H = H.at[v,:].set(neighbours_v)
        # Remove v from S and mark it
        S = S.at [v].set(False)
        M = M.at[v].set(True)
        
        if top_mode == 1:
            u = mask[:,v] & (~M) # Not marked children
            upstream_u = mask & (u[:, None]  & ~u[None,:]) # Parents of not marked children
            all_upstream_u_marked = ~jnp.any(upstream_u & ~M, axis=1)
        else:
            u = mask[v,:] & (~M) # Not marked parents
            upstream_u = mask & (u[None,:]  & ~u[:, None]) # Children of not marked parents
            all_upstream_u_marked = ~jnp.any(upstream_u & ~M, axis=1)
        
        

        S = S | (u & all_upstream_u_marked)
        S = S & (~condition_mask)

        
        return S, M, I, H
    
    _,_,_, H = jax.lax.while_loop(cond_fn, body_fn, (S, M, I, H))
    H = H | jnp.eye(num_nodes, dtype=jnp.bool_)
    H = jax.lax.cond(jnp.any(condition_mask), lambda x: x, lambda x: mask, H)
    
    # Conditioned nodes will keep the unconditional edges, hence each row of H where condition_mask is true should be equal to "mask"
    if conditioned_nodes == "unchanged":
        H = H & ~condition_mask[:, None] | mask & condition_mask[:, None]
    elif conditioned_nodes == "removed":
        H = H & ~condition_mask[:, None]
    elif conditioned_nodes == "added":
        H = H | condition_mask[:, None]
    
    return H
    
                
        
    
    
@partial(jax.jit, static_argnums=(4,))
def min_fill_heuristic(G, I, S, M, top_mode=0):
    """ Min-fill heuristic for finding a node to eliminate"""
    
    # 0 is child, 1 is parent
    UPSTREAM = top_mode
    DOWNSTREAM = 1 - top_mode
    
    # Find the number of edges that would be added if we eliminated each node
    num_edges_added = I.sum(axis=DOWNSTREAM)
    num_edges_added = S * num_edges_added + (~S) * (I.shape[0] + 1)
    # Find the node that would add the fewest edges
    # Additional constraint: Prefer marked parents
    #print(num_edges_added)
    min_val = jnp.min(num_edges_added)
    marked_parents = -jnp.sum(M[None,:] & G, axis=DOWNSTREAM)
    num_parents= marked_parents * (num_edges_added == min_val) + (I.shape[0] + 1) * (num_edges_added != min_val)
    #node_to_eliminate = jnp.argmin(num_edges_added + num_parents)
    #print(num_parents)
    reversed_array = (num_parents)[::-1]
    index = jnp.argmin(reversed_array)
    node_to_eliminate = len(reversed_array) - 1 - index
    

    return node_to_eliminate 


@jax.jit
def moralize(adj_matrix):
    adj_matrix = adj_matrix.astype(jnp.bool_)
    
    # Make the graph undirected
    undirected_graph = adj_matrix | adj_matrix.T
    
    # Add edges between parents
    undirected_graph = undirected_graph | (adj_matrix.T @ adj_matrix)
    
    return undirected_graph


def minimally_faithfull_mask(mask, condition_mask):
    """ Minimally faithfull mask update for conditioning"""
    I = moralize(mask)
    H = jnp.zeros_like(mask, dtype=jnp.bool_)
    


import jax
import jax.numpy as jnp
import jax.random as jrandom

from functools import partial


def sample_random_conditional_mask(
    key, num_samples, theta_dim, x_dim, alpha=1.0, beta=4.0
):
    # More likely to condition on a few nodes
    key1, key2 = jax.random.split(key, 2)
    condition_mask = jax.random.bernoulli(
        key1,
        jax.random.beta(key2, alpha, beta, shape=(num_samples, 1)),
        shape=(num_samples, theta_dim + x_dim),
    ).astype(jnp.bool_)
    all_ones_mask = jnp.all(condition_mask, axis=-1)
    # If all are ones, then set to false
    condition_mask = jnp.where(all_ones_mask[..., None], False, condition_mask)
    return condition_mask


def joint_conditional_mask(key, num_samples, theta_dim, x_dim):
    return jnp.array([[False] * (theta_dim + x_dim)]*num_samples)


def posterior_conditional_mask(key, num_samples, theta_dim, x_dim):
    return jnp.array([[False] * theta_dim + [True] * x_dim]*num_samples)


def likelihood_conditional_mask(key, num_samples, theta_dim, x_dim):
    return jnp.array([[True] * theta_dim + [False] * x_dim]*num_samples)


def sample_structured_conditional_mask(
    key,
    num_samples,
    theta_dim,
    x_dim,
    p_joint=0.2,
    p_posterior=0.2,
    p_likelihood=0.2,
    p_rnd1=0.2,
    p_rnd2=0.2,
    rnd1_prob=0.3,
    rnd2_prob=0.7,
):
    # Joint, posterior, likelihood, random1_mask, random2_mask
    key1, key2, key3 = jax.random.split(key, 3)
    joint_mask = jnp.array([False] * (theta_dim + x_dim), dtype=jnp.bool_)
    posterior_mask = jnp.array([False] * theta_dim + [True] * x_dim, dtype=jnp.bool_)
    likelihood_mask = jnp.array([True] * theta_dim + [False] * x_dim, dtype=jnp.bool_)
    random1_mask = jax.random.bernoulli(key2, rnd1_prob, shape=(theta_dim + x_dim,)).astype(jnp.bool_)
    random2_mask = jax.random.bernoulli(key3, rnd2_prob, shape=(theta_dim + x_dim,)).astype(jnp.bool_)
    mask_options = jnp.stack([joint_mask, posterior_mask, likelihood_mask, random1_mask, random2_mask], axis=0)  # (5, theta_dim + x_dim)
    idx = jax.random.choice(key1, 5, shape=(num_samples,), p=jnp.array([p_joint, p_posterior, p_likelihood, p_rnd1, p_rnd2]))
    condition_mask = mask_options[idx]
    all_ones_mask = jnp.all(condition_mask, axis=-1)
    # If all are ones, then set to false
    condition_mask = jnp.where(all_ones_mask[..., None], False, condition_mask)
    return condition_mask


def get_condition_mask_fn(name, **kwargs):
    if name.lower() == "structured_random":
        return partial(sample_structured_conditional_mask, **kwargs)
    elif name.lower() == "random":
        return partial(sample_random_conditional_mask, **kwargs)
    elif name.lower() == "joint":
        return partial(joint_conditional_mask, **kwargs)
    elif name.lower() == "posterior":
        return partial(posterior_conditional_mask, **kwargs)
    elif name.lower() == "likelihood":
        return partial(likelihood_conditional_mask, **kwargs)
    else:
        raise NotImplementedError()
    





import jax
from jax import numpy as jnp
import grain
import numpy as np

from datasets import load_dataset
from huggingface_hub import hf_hub_download
import json

# from .utils import download_artifacts
from .graph import faithfull_mask, min_faithfull_mask, moralize


def process_joint(batch):
    cond = batch["xs"][..., None]
    obs = batch["thetas"][..., None]
    data = np.concatenate((obs, cond), axis=1)
    return data


def process_conditional(batch):
    cond = batch["xs"][..., None]
    obs = batch["thetas"][..., None]
    return obs, cond


def normalize(batch, mean, std):
    mean = jnp.asarray(mean, dtype=batch.dtype)
    std = jnp.asarray(std, dtype=batch.dtype)
    return (batch - mean) / std


def unnormalize(batch, mean, std):
    mean = jnp.asarray(mean, dtype=batch.dtype)
    std = jnp.asarray(std, dtype=batch.dtype)
    return batch * std + mean


def has_posterior_samples(task_name):
    if task_name in [
        "two_moons",
        "bernoulli_glm",
        "gaussian_linear",
        "gaussian_linear_uniform",
        "gaussian_mixture",
        "slcp",
    ]:
        return True
    else:
        return False


class Task:
    def __init__(self, task_name, kind="joint", seed=42, use_multiprocessing=True):

        self.repo_name = "aurelio-amerio/SBI-benchmarks"

        self.task_name = task_name
        self.seed = seed
        self.use_multiprocessing = use_multiprocessing

        fname = hf_hub_download(
            repo_id=self.repo_name, filename="metadata.json", repo_type="dataset"
        )
        with open(fname, "r") as f:
            metadata = json.load(f)

        self.dataset = load_dataset(self.repo_name, task_name).with_format("numpy")

        self.df_train = self.dataset["train"]
        self.df_val = self.dataset["validation"]
        self.df_test = self.dataset["test"]

        self.max_samples = self.df_train.num_rows

        if has_posterior_samples(task_name):
            self.dataset_posterior = load_dataset(
                self.repo_name, f"{task_name}_posterior"
            ).with_format("numpy")
            self.observations = self.dataset_posterior["reference_posterior"][
                "observations"
            ]
            self.reference_samples = self.dataset_posterior["reference_posterior"][
                "reference_samples"
            ]

            self.true_parameters = self.dataset_posterior["reference_posterior"][
                "true_parameters"
            ]
            self.num_observations = len(self.observations)
        else:
            self.dataset_posterior = None
            self.observations = None
            self.reference_samples = None
            self.true_parameters = None

        self.dim_cond = metadata[task_name]["dim_cond"]
        self.dim_obs = metadata[task_name]["dim_obs"]

        if kind == "joint":
            self.dim_joint = self.dim_cond + self.dim_obs
        elif kind == "conditional":
            self.dim_joint = None
        else:
            raise ValueError(f"Unknown kind: {kind}")

        self.kind = kind

        if kind == "joint":
            self.process_fn = process_joint
        elif kind == "conditional":
            self.process_fn = process_conditional
        else:
            raise ValueError(f"Unknown kind: {kind}")

    def get_train_dataset(self, batch_size, nsamples=1e5):
        assert (
            nsamples < self.max_samples
        ), f"nsamples must be less than {self.max_samples}"

        df = self.df_train.select(range(int(nsamples)))  # [:]

        dataset_grain = (
            grain.MapDataset.source(df).shuffle(self.seed).repeat().to_iter_dataset()
        )

        performance_config = grain.experimental.pick_performance_config(
            ds=dataset_grain,
            ram_budget_mb=1024 * 4,
            max_workers=None,
            max_buffer_size=None,
        )

        dataset_batched = dataset_grain.batch(batch_size).map(self.process_fn)
        if self.use_multiprocessing:
            dataset_batched = dataset_batched.mp_prefetch(
                performance_config.multiprocessing_options
            )

        return dataset_batched

    def get_val_dataset(self, batch_size):
        df = self.df_val

        val_dataset_grain = (
            grain.MapDataset.source(df).shuffle(self.seed).repeat().to_iter_dataset()
        )
        performance_config = grain.experimental.pick_performance_config(
            ds=val_dataset_grain,
            ram_budget_mb=1024 * 4,
            max_workers=None,
            max_buffer_size=None,
        )
        val_dataset_grain = val_dataset_grain.batch(batch_size).map(self.process_fn)
        if self.use_multiprocessing:
            val_dataset_grain = val_dataset_grain.mp_prefetch(
                performance_config.multiprocessing_options
            )

        return val_dataset_grain

    def get_test_dataset(self, batch_size):
        df = self.df_test

        test_dataset_grain = (
            grain.MapDataset.source(df)
            .shuffle(self.seed)
            .repeat()
            .to_iter_dataset()
            .batch(batch_size)
            .map(self.process_fn)
        )

        return test_dataset_grain

    def get_reference(self, num_observation=1):
        """
        Returns the reference posterior samples for a given number of observations.
        """
        if num_observation < 1 or num_observation > self.num_observations:
            raise ValueError(
                f"num_observation must be between 1 and {self.num_observations}"
            )
        obs = self.observations[num_observation - 1]
        samples = self.reference_samples[num_observation - 1]
        return obs, samples

    def get_true_parameters(self, num_observation=1):
        """
        Returns the true parameters for a given number of observations.
        """
        if num_observation < 1 or num_observation > self.num_observations:
            raise ValueError(
                f"num_observation must be between 1 and {self.num_observations}"
            )
        return self.true_parameters[num_observation - 1]

    def get_base_mask_fn(self):
        raise NotImplementedError()

    def get_edge_mask_fn(self, name="undirected"):
        if name.lower() == "faithfull":
            base_mask_fn = self.get_base_mask_fn()

            def faithfull_edge_mask(node_id, condition_mask, meta_data=None):
                base_mask = base_mask_fn(node_id, meta_data)
                return faithfull_mask(base_mask, condition_mask)

            return faithfull_edge_mask
        elif name.lower() == "min_faithfull":
            base_mask_fn = self.get_base_mask_fn()

            def min_faithfull_edge_mask(node_id, condition_mask, meta_data=None):
                base_mask = base_mask_fn(node_id, meta_data)

                return min_faithfull_mask(base_mask, condition_mask)

            return min_faithfull_edge_mask
        elif name.lower() == "undirected":
            base_mask_fn = self.get_base_mask_fn()

            def undirected_edge_mask(node_id, condition_mask, meta_data=None):
                base_mask = base_mask_fn(node_id, meta_data)
                return moralize(base_mask)

            return undirected_edge_mask

        elif name.lower() == "directed":
            base_mask_fn = self.get_base_mask_fn()

            def directed_edge_mask(node_id, condition_mask, meta_data=None):
                base_mask = base_mask_fn(node_id, meta_data)
                return base_mask

            return directed_edge_mask
        elif name.lower() == "none":
            return lambda node_id, condition_mask, *args, **kwargs: None
        else:
            raise NotImplementedError()


class TwoMoons(Task):
    def __init__(self, kind="joint", **kwargs):
        task_name = "two_moons"
        super().__init__(task_name, kind=kind, **kwargs)

    def get_base_mask_fn(self):
        theta_dim = self.dim_obs
        x_dim = self.dim_cond
        thetas_mask = jnp.eye(theta_dim, dtype=jnp.bool_)
        x_mask = jnp.tril(jnp.ones((theta_dim, x_dim), dtype=jnp.bool_))
        base_mask = jnp.block(
            [
                [thetas_mask, jnp.zeros((theta_dim, x_dim))],
                [jnp.ones((x_dim, theta_dim)), x_mask],
            ]
        )
        base_mask = base_mask.astype(jnp.bool_)

        def base_mask_fn(node_ids, node_meta_data):
            return base_mask[node_ids, :][:, node_ids]

        return base_mask_fn


class BernoulliGLM(Task):
    def __init__(self, kind="joint", **kwargs):
        task_name = "bernoulli_glm"
        super().__init__(task_name, kind=kind, **kwargs)

    def get_base_mask_fn(self):
        raise NotImplementedError()


class GaussianLinear(Task):
    def __init__(self, kind="joint", **kwargs):
        task_name = "gaussian_linear"
        super().__init__(task_name, kind=kind, **kwargs)

    def get_base_mask_fn(self):
        theta_dim = self.dim_obs
        x_dim = self.dim_cond
        thetas_mask = jnp.eye(theta_dim, dtype=jnp.bool_)
        x_i_mask = jnp.eye(x_dim, dtype=jnp.bool_)
        base_mask = jnp.block(
            [[thetas_mask, jnp.zeros((theta_dim, x_dim))], [jnp.eye((x_dim)), x_i_mask]]
        )
        base_mask = base_mask.astype(jnp.bool_)

        def base_mask_fn(node_ids, node_meta_data):
            return base_mask[node_ids, :][:, node_ids]

        return base_mask_fn


class GaussianLinearUniform(Task):
    def __init__(self, kind="joint", **kwargs):
        task_name = "gaussian_linear_uniform"
        super().__init__(task_name, kind=kind, **kwargs)

    def get_base_mask_fn(self):
        theta_dim = self.dim_obs
        x_dim = self.dim_cond
        thetas_mask = jnp.eye(theta_dim, dtype=jnp.bool_)
        x_i_mask = jnp.eye(x_dim, dtype=jnp.bool_)
        base_mask = jnp.block(
            [[thetas_mask, jnp.zeros((theta_dim, x_dim))], [jnp.eye((x_dim)), x_i_mask]]
        )
        base_mask = base_mask.astype(jnp.bool_)

        def base_mask_fn(node_ids, node_meta_data):
            return base_mask[node_ids, :][:, node_ids]

        return base_mask_fn


class GaussianMixture(Task):
    def __init__(self, kind="joint", **kwargs):
        task_name = "gaussian_mixture"
        super().__init__(task_name, kind=kind, **kwargs)

    def get_base_mask_fn(self):
        theta_dim = self.dim_obs
        x_dim = self.dim_cond
        thetas_mask = jnp.eye(theta_dim, dtype=jnp.bool_)
        x_mask = jnp.tril(jnp.ones((theta_dim, x_dim), dtype=jnp.bool_))
        base_mask = jnp.block(
            [
                [thetas_mask, jnp.zeros((theta_dim, x_dim))],
                [jnp.ones((x_dim, theta_dim)), x_mask],
            ]
        )
        base_mask = base_mask.astype(jnp.bool_)

        def base_mask_fn(node_ids, node_meta_data):
            return base_mask[node_ids, :][:, node_ids]

        return base_mask_fn


class SLCP(Task):
    def __init__(self, kind="joint", **kwargs):
        task_name = "slcp"
        super().__init__(task_name, kind=kind, **kwargs)

    def get_base_mask_fn(self):
        theta_dim = self.dim_obs
        x_dim = self.dim_cond
        thetas_mask = jnp.eye(theta_dim, dtype=jnp.bool_)
        x_i_dim = x_dim // 4
        x_i_mask = jax.scipy.linalg.block_diag(
            *tuple([jnp.tril(jnp.ones((x_i_dim, x_i_dim), dtype=jnp.bool_))] * 4)
        )
        base_mask = jnp.block(
            [
                [thetas_mask, jnp.zeros((theta_dim, x_dim))],
                [jnp.ones((x_dim, theta_dim)), x_i_mask],
            ]
        )
        base_mask = base_mask.astype(jnp.bool_)

        def base_mask_fn(node_ids, node_meta_data):
            return base_mask[node_ids, :][:, node_ids]

        return base_mask_fn


class GravitationalWaves(Task):
    def __init__(self, **kwargs):
        task_name = "gravitational_waves"
        super().__init__(task_name, kind="conditional", **kwargs)

        self.dim_obs = 2
        self.ch_obs = 1
        dim_cond_tot = self.dim_cond  # from super
        self.dim_cond = 8192
        self.ch_cond = 2
        assert (
            self.dim_cond == dim_cond_tot[0] and self.ch_cond == dim_cond_tot[1]
        ), f"Dimension mismatch, expected ({dim_cond_tot[0]}, {dim_cond_tot[1]}), got ({self.dim_cond}, {self.ch_cond})"

        self.xs_mean = jnp.array([[[0.00051776, -0.00040733]]], dtype=jnp.bfloat16)
        self.thetas_mean = jnp.array([[44.826576, 45.070328]], dtype=jnp.bfloat16)

        self.xs_std = jnp.array([[[60.80799, 59.33193]]], dtype=jnp.bfloat16)
        self.thetas_std = jnp.array([[20.189356, 20.16127]], dtype=jnp.bfloat16)

        self.process_fn = self.split_data

        return

    def split_data(self, batch):
        obs = jnp.array(batch["thetas"], dtype=jnp.bfloat16)
        obs = normalize(obs, self.thetas_mean, self.thetas_std)
        obs = obs.reshape(obs.shape[0], self.dim_obs, self.ch_obs)
        cond = jnp.array(batch["xs"], dtype=jnp.bfloat16)
        cond = normalize(cond, self.xs_mean, self.xs_std)
        cond = cond[..., None]
        return obs, cond

    def get_reference(self, num_observation=1):
        raise NotImplementedError(
            "Reference posterior samples not available for this task."
        )

    def get_true_parameters(self, num_observation=1):
        raise NotImplementedError("True parameters not available for this task.")


class GravitationalLensing(Task):
    def __init__(self, **kwargs):
        task_name = "lensing"
        super().__init__(task_name, kind="conditional", **kwargs)

        self.dim_obs = 2
        self.ch_obs = 1
        dim_cond_tot = self.dim_cond  # from super
        self.dim_cond = 32
        self.ch_cond = 32
        assert (
            self.dim_cond == dim_cond_tot[0] and self.ch_cond == dim_cond_tot[1]
        ), f"Dimension mismatch, expected ({dim_cond_tot[0]}, {dim_cond_tot[1]}), got ({self.dim_cond}, {self.ch_cond})"

        self.xs_mean = jnp.array([-1.1874731e-05], dtype=jnp.bfloat16).reshape(1, 1, 1)
        self.thetas_mean = jnp.array(
            [0.5996428, 0.15998043], dtype=jnp.bfloat16
        ).reshape(1, 2)

        self.xs_std = jnp.array([1.0440514], dtype=jnp.bfloat16).reshape(1, 1, 1)
        self.thetas_std = jnp.array(
            [0.2886958, 0.08657552], dtype=jnp.bfloat16
        ).reshape(1, 2)

        self.process_fn = self.split_data
        return

    def split_data(self, batch):
        obs = jnp.array(batch["thetas"], dtype=jnp.bfloat16)
        obs = normalize(obs, self.thetas_mean, self.thetas_std)
        obs = obs.reshape(obs.shape[0], self.dim_obs, self.ch_obs)
        cond = jnp.array(batch["xs"], dtype=jnp.bfloat16)
        cond = normalize(cond, self.xs_mean, self.xs_std)
        cond = cond[..., None]
        return obs, cond

    def get_reference(self, num_observation=1):
        raise NotImplementedError(
            "Reference posterior samples not available for this task."
        )

    def get_true_parameters(self, num_observation=1):
        raise NotImplementedError("True parameters not available for this task.")


def get_task(task_name, kind="conditional", **kwargs):
    """
    Returns a Task object based on the task name.
    """
    task_name = task_name.lower()
    if task_name == "two_moons":
        return TwoMoons(kind=kind, **kwargs)
    elif task_name == "bernoulli_glm":
        return BernoulliGLM(kind=kind, **kwargs)
    elif task_name == "gaussian_linear":
        return GaussianLinear(kind=kind, **kwargs)
    elif task_name == "gaussian_linear_uniform":
        return GaussianLinearUniform(kind=kind, **kwargs)
    elif task_name == "gaussian_mixture":
        return GaussianMixture(kind=kind, **kwargs)
    elif task_name == "slcp":
        return SLCP(kind=kind, **kwargs)
    elif task_name == "gravitational_waves":
        assert (
            kind == "conditional"
        ), "Gravitational waves task is only available in conditional mode."
        return GravitationalWaves(**kwargs)
    elif task_name == "gravitational_lensing":
        assert (
            kind == "conditional"
        ), "Gravitational lensing task is only available in conditional mode."
        return GravitationalLensing(**kwargs)
    else:
        raise ValueError(f"Unknown task: {task_name}")
